from ..py_object import PyObject


class VideoParameters(PyObject):
    """Video parameters of the stream

    Attributes:
        width (``int``):
            Video width
        height (``int``):
            Video height
        frame_rate (``int``):
            Framerate of video

    Parameters:
        width (``int``):
            Video width
        height (``int``):
            Video height
        frame_rate (``int``):
            Framerate of video
    """

    def __init__(
        self,
        width: int = 640,
        height: int = 360,
        frame_rate: int = 20,
    ):
        self.width: int = width
        self.height: int = height
        self.frame_rate: int = frame_rate



from ...scaffold import Scaffold
from ...types.call_holder import CallHolder
from ...types.stream import StreamAudioEnded
from ...types.stream import StreamVideoEnded


class StreamEndedHandler(Scaffold):
    async def _stream_ended_handler(
        self,
        params: dict,
        is_audio: bool,
    ):
        chat_id = int(params['chat_id'])
        self._call_holder.set_status(
            chat_id,
            CallHolder.IDLE,
        )
        await self._on_event_update.propagate(
            'STREAM_END_HANDLER',
            self,
            StreamAudioEnded(
                chat_id,
            ) if is_audio else StreamVideoEnded(chat_id),
        )
        return {
            'result': 'OK',
        }



import json
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import fipper
from fipper import Client
from fipper import ContinuePropagation
from fipper.raw.base import InputPeer
from fipper.raw.functions.channels import GetFullChannel
from fipper.raw.functions.messages import GetFullChat
from fipper.raw.functions.phone import EditGroupCallParticipant
from fipper.raw.functions.phone import GetGroupCall
from fipper.raw.functions.phone import GetGroupParticipants
from fipper.raw.functions.phone import JoinGroupCall
from fipper.raw.functions.phone import LeaveGroupCall
from fipper.raw.types import Channel
from fipper.raw.types import ChannelForbidden
from fipper.raw.types import Chat
from fipper.raw.types import ChatForbidden
from fipper.raw.types import DataJSON
from fipper.raw.types import GroupCall
from fipper.raw.types import GroupCallDiscarded
from fipper.raw.types import InputChannel
from fipper.raw.types import InputGroupCall
from fipper.raw.types import InputPeerChannel
from fipper.raw.types import MessageActionChatDeleteUser
from fipper.raw.types import MessageActionInviteToGroupCall
from fipper.raw.types import MessageService
from fipper.raw.types import PeerChat
from fipper.raw.types import UpdateChannel
from fipper.raw.types import UpdateGroupCall
from fipper.raw.types import UpdateGroupCallConnection
from fipper.raw.types import UpdateGroupCallParticipants
from fipper.raw.types import UpdateNewChannelMessage
from fipper.raw.types import UpdateNewMessage
from fipper.raw.types import Updates

from ..version_manager import VersionManager
from .bridged_client import BridgedClient
from .client_cache import ClientCache


class FipperClient(BridgedClient):
    def __init__(
        self,
        cache_duration: int,
        client: Client,
    ):
        self._app: Client = client
        if VersionManager.version_tuple(
            fipper.__version__,
        ) > VersionManager.version_tuple(
            '2.0.0',
        ):
            self._app.send = self._app.invoke
        self._handler: Dict[str, Callable] = {}
        self._cache: ClientCache = ClientCache(
            cache_duration,
            self,
        )

        @self._app.on_raw_update()
        async def on_update(_, update, __, data2):
            if isinstance(
                update,
                UpdateGroupCallParticipants,
            ):
                participants = update.participants
                for participant in participants:
                    result = self._cache.set_participants_cache(
                        update.call.id,
                        self.chat_id(participant.peer),
                        participant.muted,
                        participant.volume,
                        participant.can_self_unmute,
                        participant.video is not None or
                        participant.presentation is not None,
                        participant.presentation is not None,
                        participant.video is not None,
                        participant.raise_hand_rating,
                        participant.left,
                    )
                    if result is not None:
                        if 'PARTICIPANTS_HANDLER' in self._handler:
                            await self._handler['PARTICIPANTS_HANDLER'](
                                self._cache.get_chat_id(update.call.id),
                                result,
                                participant.just_joined,
                                participant.left,
                            )
            if isinstance(
                update,
                UpdateGroupCall,
            ):
                chat_id = self.chat_id(data2[update.chat_id])
                if isinstance(
                    update.call,
                    GroupCall,
                ):
                    if update.call.schedule_date is None:
                        self._cache.set_cache(
                            chat_id,
                            InputGroupCall(
                                access_hash=update.call.access_hash,
                                id=update.call.id,
                            ),
                        )
                if isinstance(
                    update.call,
                    GroupCallDiscarded,
                ):
                    self._cache.drop_cache(chat_id)
                    if 'CLOSED_HANDLER' in self._handler:
                        await self._handler['CLOSED_HANDLER'](
                            chat_id,
                        )
            if isinstance(
                update,
                UpdateChannel,
            ):
                chat_id = self.chat_id(update)
                if len(data2) > 0:
                    if isinstance(
                        data2[update.channel_id],
                        ChannelForbidden,
                    ):
                        self._cache.drop_cache(chat_id)
                        if 'KICK_HANDLER' in self._handler:
                            await self._handler['KICK_HANDLER'](
                                chat_id,
                            )
            if isinstance(
                update,
                UpdateNewChannelMessage,
            ) or isinstance(
                update,
                UpdateNewMessage,
            ):
                if isinstance(
                    update.message,
                    MessageService,
                ):
                    if isinstance(
                        update.message.action,
                        MessageActionInviteToGroupCall,
                    ):
                        if 'INVITE_HANDLER' in self._handler:
                            await self._handler['INVITE_HANDLER'](
                                update.message.action,
                            )
                    if isinstance(
                        update.message.action,
                        MessageActionChatDeleteUser,
                    ):
                        if isinstance(
                            update.message.peer_id,
                            PeerChat,
                        ):
                            chat_id = self.chat_id(update.message.peer_id)
                            if isinstance(
                                data2[update.message.peer_id.chat_id],
                                ChatForbidden,
                            ):
                                self._cache.drop_cache(chat_id)
                                if 'KICK_HANDLER' in self._handler:
                                    await self._handler['KICK_HANDLER'](
                                        chat_id,
                                    )
            if isinstance(
                data2,
                Dict,
            ):
                for group_id in data2:
                    if isinstance(
                        update,
                        UpdateNewChannelMessage,
                    ) or isinstance(
                        update,
                        UpdateNewMessage,
                    ):
                        if isinstance(
                            update.message,
                            MessageService,
                        ):
                            if isinstance(
                                data2[group_id],
                                Channel,
                            ) or isinstance(
                                data2[group_id],
                                Chat,
                            ):
                                chat_id = self.chat_id(data2[group_id])
                                if data2[group_id].left:
                                    self._cache.drop_cache(
                                        chat_id,
                                    )
                                    if 'LEFT_HANDLER' in self._handler:
                                        await self._handler['LEFT_HANDLER'](
                                            chat_id,
                                        )
            raise ContinuePropagation()

    def on_closed_voice_chat(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._handler['CLOSED_HANDLER'] = func
            return func
        return decorator

    def on_kicked(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._handler['KICK_HANDLER'] = func
            return func
        return decorator

    def on_receive_invite(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._handler['INVITE_HANDLER'] = func
            return func
        return decorator

    def on_left_group(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._handler['LEFT_HANDLER'] = func
            return func
        return decorator

    def on_participants_change(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._handler['PARTICIPANTS_HANDLER'] = func
            return func
        return decorator

    async def get_call(
        self,
        chat_id: int,
    ) -> Optional[InputGroupCall]:
        chat = await self._app.resolve_peer(chat_id)
        if isinstance(chat, InputPeerChannel):
            input_call = (
                await self._app.send(
                    GetFullChannel(
                        channel=InputChannel(
                            channel_id=chat.channel_id,
                            access_hash=chat.access_hash,
                        ),
                    ),
                )
            ).full_chat.call
        else:
            input_call = (
                await self._app.send(
                    GetFullChat(chat_id=chat.chat_id),
                )
            ).full_chat.call
        if input_call is not None:
            call: GroupCall = (
                await self._app.send(
                    GetGroupCall(
                        call=input_call,
                        limit=-1,
                    ),
                )
            ).call
            if call.schedule_date is not None:
                return None
        return input_call

    async def get_group_call_participants(
        self,
        chat_id: int,
    ):
        return await self._cache.get_participant_list(
            chat_id,
        )

    async def get_participants(
        self,
        input_call: InputGroupCall,
    ):
        return [
            {
                'user_id': self.chat_id(participant.peer),
                'muted': participant.muted,
                'volume': participant.volume,
                'can_self_unmute': participant.can_self_unmute,
                'video': participant.video,
                'presentation': participant.presentation,
                'raise_hand_rating': participant.raise_hand_rating,
                'left': participant.left,
            } for participant in (
                await self._app.send(
                    GetGroupParticipants(
                        call=input_call,
                        ids=[],
                        sources=[],
                        offset='',
                        limit=500,
                    ),
                )
            ).participants
        ]

    async def join_group_call(
        self,
        chat_id: int,
        json_join: dict,
        invite_hash: str,
        have_video: bool,
        join_as: InputPeer,
    ) -> dict:
        chat_call = await self._cache.get_full_chat(chat_id)
        if chat_call is not None:
            result: Updates = await self._app.send(
                JoinGroupCall(
                    call=chat_call,
                    params=DataJSON(data=json.dumps(json_join)),
                    muted=False,
                    join_as=join_as,
                    video_stopped=have_video,
                    invite_hash=invite_hash,
                ),
            )
            for update in result.updates:
                if isinstance(
                    update,
                    UpdateGroupCallParticipants,
                ):
                    participants = update.participants
                    for participant in participants:
                        self._cache.set_participants_cache(
                            update.call.id,
                            self.chat_id(participant.peer),
                            participant.muted,
                            participant.volume,
                            participant.can_self_unmute,
                            participant.video is not None or
                            participant.presentation is not None,
                            participant.presentation is not None,
                            participant.video is not None,
                            participant.raise_hand_rating,
                            participant.left,
                        )
                if isinstance(update, UpdateGroupCallConnection):
                    data_json = json.loads(update.params.data)
                    if 'rtmp' in data_json:
                        raise Exception('APP_UPGRADE_NEEDED')
                    elif 'transport' not in data_json:
                        raise Exception('No transport in update')
                    transport = data_json['transport']
                    return {
                        'transport': {
                            'ufrag': transport['ufrag'],
                            'pwd': transport['pwd'],
                            'fingerprints': transport['fingerprints'],
                            'candidates': transport['candidates'],
                        },
                    }
        return {'transport': None}

    async def leave_group_call(
        self,
        chat_id: int,
    ):
        chat_call = await self._cache.get_full_chat(chat_id)
        if chat_call is not None:
            await self._app.send(
                LeaveGroupCall(
                    call=chat_call,
                    source=0,
                ),
            )

    async def change_volume(
        self,
        chat_id: int,
        volume: int,
        participant: InputPeer,
    ):
        chat_call = await self._cache.get_full_chat(chat_id)
        if chat_call is not None:
            await self._app.send(
                EditGroupCallParticipant(
                    call=chat_call,
                    participant=participant,
                    muted=False,
                    volume=volume * 100,
                ),
            )

    async def set_video_call_status(
        self,
        chat_id: int,
        stopped_status: Optional[bool],
        paused_status: Optional[bool],
        participant: InputPeer,
    ):
        chat_call = await self._cache.get_full_chat(chat_id)
        if chat_call is not None:
            await self._app.send(
                EditGroupCallParticipant(
                    call=chat_call,
                    participant=participant,
                    muted=False,
                    video_stopped=stopped_status,
                    video_paused=paused_status,
                ),
            )

    async def get_full_chat(self, chat_id: int):
        return await self._cache.get_full_chat(chat_id)

    async def resolve_peer(
        self,
        user_id: Union[int, str],
    ) -> InputPeer:
        return await self._app.resolve_peer(user_id)

    async def get_id(self) -> int:
        return (await self._app.get_me()).id

    def is_connected(self) -> bool:
        return self._app.is_connected

    async def start(self):
        await self._app.start()



from asyncio import Future
from typing import Optional


class Scaffold:
    _REQUIRED_NODEJS_VERSION = '15.0.0'
    _REQUIRED_FIPPER_VERSION = '0.0.1.dev2'
    _REQUIRED_TELETHON_VERSION = '1.24.0'

    def __init__(self):
        self._app = None
        self._async_core = None
        self._is_running = None
        self._my_id = None
        self._wait_until_run: Optional[Future] = None
        self._env_checker = None
        self._call_holder = None
        self._cache_user_peer = None
        self._wait_result = None
        self._cache_local_peer = None
        self._on_event_update = None
        self._binding = None

    def _handle_mtproto(self):
        pass

    async def _start_binding(self):
        pass

    async def _init_mtproto(self):
        pass

    async def _join_voice_call(self, params: dict):
        pass

    async def _leave_voice_call(self, params: dict):
        pass

    async def _stream_ended_handler(self, params: dict, is_audio: bool):
        pass

    async def _raw_update_handler(self, params: dict):
        pass

    async def _set_video_call_status(self, params: dict):
        pass

    async def start(self):
        pass



import atexit
from typing import Any

from .binding import Binding
from .environment import Environment
from .handlers import HandlersHolder
from .methods import Methods
from .mtproto import MtProtoClient
from .scaffold import Scaffold
from .types import Cache
from .types.call_holder import CallHolder
from .types.update_solver import UpdateSolver


class PyCover(Methods, Scaffold):
    """PyCover Client, the main means
    for interacting with Group Calls.

    Attributes:
        active_calls (List of :obj:`~pycover.types.GroupCall`):
            Get a list of active (Playing / Paused) group calls
        calls (List of :obj:`~pycover.types.GroupCall`):
            Get a list of existent group calls
        cache_peer (`InputPeer (P)`_ | `InputPeer (T)`_):
            Get current Telegram user
        ping (``int``):
            Ping of NodeJS core
        is_connected (``bool``):
            Check if is alive the NodeJS connection

    Parameters:
        app (`Client`_ | `TelegramClient`_):
            Pass the MtProto Client

        cache_duration (``int``):
            Cache duration of Full Chat query

        overload_quiet_mode (``bool``):
            Disable overload cpu messages by setting true

    Raises:
        InvalidMtProtoClient: You set an invalid MtProto client

    """

    def __init__(
        self,
        app: Any,
        cache_duration: int = 120,
        overload_quiet_mode: bool = False,
    ):
        super().__init__()
        self._app = MtProtoClient(
            cache_duration,
            app,
        )
        self._is_running = False
        self._env_checker = Environment(
            self._REQUIRED_NODEJS_VERSION,
            self._REQUIRED_FIPPER_VERSION,
            self._REQUIRED_TELETHON_VERSION,
            self._app.client,
        )
        self._call_holder = CallHolder()
        self._cache_user_peer = Cache()
        self._wait_result = UpdateSolver()
        self._on_event_update = HandlersHolder()
        self._binding = Binding(
            overload_quiet_mode,
        )

        def cleanup():
            if self._async_core is not None:
                self._async_core.cancel()
        atexit.register(cleanup)



from ...ffprobe import FFprobe
from ...media_devices.device_info import DeviceInfo
from ...media_devices.screen_info import ScreenInfo
from .audio_parameters import AudioParameters
from .input_stream import InputAudioStream
from .input_stream import InputStream
from .input_stream import InputVideoStream
from .video_parameters import VideoParameters


class CaptureAVDeviceDesktop(InputStream):
    """Capture video from Screen and Audio from device

    Attributes:
        stream_audio (:obj:`~pycover.types.InputAudioStream()`):
            Input Audio Stream Descriptor
        stream_video (:obj:`~pycover.types.InputVideoStream()`):
            Input Video Stream Descriptor
    Parameters:
        audio_info (:obj: `~pycover.media_devices.DeviceInfo()`):
            The audio device capturing params
        screen_info (:obj: `~pycover.media_devices.ScreenManager()`):
            The screen video capturing params
        audio_parameters (:obj:`~pycover.types.AudioParameters()`):
            The audio parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityAudio()`,
            :obj:`~pycover.types.MediumQualityAudio()` or
            :obj:`~pycover.types.LowQualityAudio()`
        video_parameters (:obj:`~pycover.types.VideoParameters()`):
            The video parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityVideo()`,
            :obj:`~pycover.types.MediumQualityVideo()` or
            :obj:`~pycover.types.LowQualityVideo()`
    """

    def __init__(
        self,
        audio_info: DeviceInfo,
        screen_info: ScreenInfo,
        audio_parameters: AudioParameters = AudioParameters(),
        video_parameters: VideoParameters = VideoParameters(),
    ):
        self._audio_path = audio_info.build_ffmpeg_command()
        self.audio_ffmpeg: str = audio_info.ffmpeg_parameters
        self._video_path = screen_info.build_ffmpeg_command(
            video_parameters.frame_rate,
        )
        self.video_ffmpeg: str = screen_info.ffmpeg_parameters
        self.raw_headers = None
        super().__init__(
            InputAudioStream(
                f'device://{self._audio_path}',
                audio_parameters,
            ),
            InputVideoStream(
                f'screen://{self._video_path}',
                video_parameters,
            ),
        )

    @property
    def headers(self):
        return FFprobe.ffmpeg_headers(self.raw_headers)

    async def check_pipe(self):
        pass



from ..py_object import PyObject
from .audio_parameters import AudioParameters


class InputAudioStream(PyObject):
    """The raw audio stream (PCM16L) descriptor

    Attributes:
        path (``str``):
            The audio file path
        parameters (:obj:`~pycover.types.AudioParameters()`):
            The audio parameters of the stream

    Parameters:
        path (``str``):
            The audio file path (PCM16L format only)
        parameters (:obj:`~pycover.types.AudioParameters()`):
            The audio parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityAudio()`,
            :obj:`~pycover.types.MediumQualityAudio()` or
            :obj:`~pycover.types.LowQualityAudio()`
    """

    def __init__(
        self,
        path: str,
        parameters: AudioParameters = AudioParameters(),
        header_enabled: bool = False,
    ):
        self.path: str = path
        self.parameters: AudioParameters = parameters
        self.header_enabled: bool = header_enabled



import asyncio
from typing import Union

from ...exceptions import NodeJSNotRunning
from ...exceptions import NoMtProtoClientSet
from ...exceptions import NotInGroupCallError
from ...mtproto import BridgedClient
from ...scaffold import Scaffold
from ...types import NotInGroupCall
from ...types.session import Session
from ...types.stream import StreamTime


class PlayedTime(Scaffold):
    async def played_time(
        self,
        chat_id: Union[int, str],
    ):
        """Get the played time of the stream

        This method allow to get the played time of the stream

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier of the target chat.
                Can be a direct id (int) or a username (str)

        Raises:
            NoMtProtoClientSet: In case you try
                to call this method without any MtProto client
            NodeJSNotRunning: In case you try
                to call this method without do
                :meth:`~pycover.pycover.start` before
            NoActiveGroupCall: In case you try
                to edit a not started group call
        """
        try:
            chat_id = int(chat_id)
        except ValueError:
            chat_id = BridgedClient.chat_id(
                await self._app.resolve_peer(chat_id),
            )
        if self._app is not None:
            if self._wait_until_run is not None:
                solver_id = Session.generate_session_id(24)

                async def internal_sender():
                    if not self._wait_until_run.done():
                        await self._wait_until_run
                    request = {
                        'action': 'played_time',
                        'chat_id': chat_id,
                        'solver_id': solver_id,
                    }
                    await self._binding.send(request)

                asyncio.ensure_future(internal_sender())
                result = await self._wait_result.wait_future_update(
                    solver_id,
                )
                if isinstance(result, NotInGroupCall):
                    raise NotInGroupCallError()
                elif isinstance(result, StreamTime):
                    return result.time
            else:
                raise NodeJSNotRunning()
        else:
            raise NoMtProtoClientSet()



from ...types.update import Update


class MutedStream(Update):
    """Raised when muted stream successfully

    Attributes:
        chat_id (``int``):
            Unique identifier of chat.

    Parameters:
        chat_id (``int``):
            Unique identifier of chat.
    """

    def __init__(
        self,
        chat_id: int,
    ):
        super().__init__(chat_id)



from ..video_parameters import VideoParameters


class MediumQualityVideo(VideoParameters):
    """Medium Video Quality (854x480)

    Attributes:
        width (``int``):
            Video width
        height (``int``):
            Video height
        frame_rate (``int``):
            Framerate of video
    """

    def __init__(self):
        super().__init__(
            854,
            480,
            20,
        )



import os

from .exceptions import NodeJSNotInstalled
from .exceptions import TooOldNodeJSVersion
from .exceptions import TooOldFipperVersion
from .exceptions import TooOldTelethonVersion
from .version_manager import VersionManager


class Environment:
    def __init__(
        self,
        min_js_version: str,
        min_fipper_version: str,
        min_telethon_version: str,
        client_name: str,
    ):
        self._REQUIRED_NODEJS_VERSION = min_js_version
        self._REQUIRED_FIPPER_VERSION = min_fipper_version
        self._REQUIRED_TELETHON_VERSION = min_telethon_version
        self._client_name = client_name

    def check_environment(self):
        def get_version(package_check):
            result_cmd = os.popen(f'{package_check} -v').read()
            result_cmd = result_cmd.replace('v', '')
            if len(result_cmd) == 0:
                return None
            return result_cmd

        node_result = get_version('node')
        if node_result is None:
            raise NodeJSNotInstalled(
                self._REQUIRED_NODEJS_VERSION,
            )
        if VersionManager.version_tuple(
            node_result,
        ) < VersionManager.version_tuple(
            self._REQUIRED_NODEJS_VERSION,
        ):
            raise TooOldNodeJSVersion(
                self._REQUIRED_NODEJS_VERSION,
                node_result,
            )
        if self._client_name == 'fipper':
            import fipper
            if VersionManager.version_tuple(
                fipper.__version__,
            ) < VersionManager.version_tuple(
                self._REQUIRED_FIPPER_VERSION,
            ):
                raise TooOldFipperVersion(
                    self._REQUIRED_FIPPER_VERSION,
                    fipper.__version__,
                )
        elif self._client_name == 'telethon':
            import telethon
            if VersionManager.version_tuple(
                telethon.__version__,
            ) < VersionManager.version_tuple(
                self._REQUIRED_TELETHON_VERSION,
            ):
                raise TooOldTelethonVersion(
                    self._REQUIRED_TELETHON_VERSION,
                    telethon.__version__,
                )



import asyncio
import json
import subprocess
from json import JSONDecodeError
from typing import Dict
from typing import List
from typing import Optional

from .exceptions import FFmpegNotInstalled
from .exceptions import InvalidVideoProportion
from .exceptions import NoAudioSourceFound
from .exceptions import NoVideoSourceFound
from .types.input_stream.video_tools import check_support


class FFprobe:
    @staticmethod
    def ffmpeg_headers(
        headers: Optional[Dict[str, str]] = None,
    ):
        ffmpeg_params: List[str] = []
        if headers is not None:
            ffmpeg_params.append('-headers')
            built_header = ''
            for i in headers:
                built_header += f'{i}: {headers[i]}\r\n'
            ffmpeg_params.append(built_header)
        return ':_cmd_:'.join(
            ffmpeg_params,
        )

    @staticmethod
    async def check_file(
        path: str,
        needed_audio=False,
        needed_video=False,
        needed_image=False,
        headers: Optional[Dict[str, str]] = None,
    ):
        ffmpeg_params: List[str] = []
        have_header = False
        if headers is not None and \
                check_support(path):
            ffmpeg_params.append('-headers')
            built_header = ''
            have_header = True
            for i in headers:
                built_header += f'{i}: {headers[i]}\r\n'
            ffmpeg_params.append(built_header)
        try:
            ffprobe = await asyncio.create_subprocess_exec(
                'ffprobe',
                '-v',
                'error',
                '-show_entries',
                'stream=width,height,codec_type,codec_name',
                '-of',
                'json',
                path,
                *tuple(ffmpeg_params),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stream_list = []
            try:
                stdout, stderr = await asyncio.wait_for(
                    ffprobe.communicate(),
                    timeout=30,
                )
                result = json.loads(stdout.decode('utf-8')) or {}
                stream_list = result.get('streams', [])
            except (subprocess.TimeoutExpired, JSONDecodeError):
                pass
            have_video = False
            have_audio = False
            have_valid_video = False
            original_width = 0
            original_height = 0
            for stream in stream_list:
                codec_type = stream.get('codec_type', '')
                codec_name = stream.get('codec_name', '')
                image_codecs = ['png', 'jpeg', 'jpg', 'mjpeg']
                is_valid = not needed_image and codec_name in image_codecs
                if codec_type == 'video' and not is_valid:
                    have_video = True
                    original_width = int(stream.get('width', 0))
                    original_height = int(stream.get('height', 0))
                    if original_height and original_width:
                        have_valid_video = True
                elif codec_type == 'audio':
                    have_audio = True
            if needed_video:
                if not have_video:
                    raise NoVideoSourceFound(path)
                if not have_valid_video:
                    raise InvalidVideoProportion(
                        'Video proportion not found',
                    )
            if needed_audio:
                if not have_audio:
                    raise NoAudioSourceFound(path)
                if not needed_video:
                    return have_header
            if have_video:
                return original_width, original_height, have_header
        except FileNotFoundError:
            raise FFmpegNotInstalled(path)



from pycover.types.py_object import PyObject


class GroupCall(PyObject):
    """Info about a group call

    Attributes:
        chat_id (``int``):
            Unique identifier of chat.
        is_playing (``bool``):
            Check if exist any sort of stream
        status (``str``):
            Status of Stream

    Parameters:
        chat_id (``int``):
            Unique identifier of chat.
        binary_status (``int``):
            PyCover API parameter.
    """

    def __init__(
            self,
            chat_id: int,
            binary_status: int,
    ):
        self.chat_id: int = chat_id
        self.is_playing: bool = binary_status != 3
        self.status: str = 'unknown'
        if binary_status == 1:
            self.status = 'playing'
        elif binary_status == 2:
            self.status = 'paused'
        elif binary_status == 3:
            self.status = 'not_playing'



class StreamTime:
    """Played time of the stream

    Attributes:
        time (``int``):
            Time of the stream in seconds.

    Parameters:
        time (``int``):
            Time of the stream in seconds.
    """

    def __init__(
        self,
        time: int,
    ):
        self.time = time



from ..audio_parameters import AudioParameters


class HighQualityAudio(AudioParameters):
    """High Audio Quality (48K of bitrate)

    Attributes:
        bitrate (``int``):
            Audio bitrate
    """

    def __init__(self):
        super().__init__(
            48000,
        )



class NodeJSNotInstalled(Exception):
    """Node.js isn’t installed, raised by
    :meth:`~pycover.pycover.start` or
    :meth:`~pycover.pycover.run`
    """

    def __init__(
        self,
        version_needed: str,
    ):
        super().__init__(
            f'Please install node ({version_needed}+)',
        )


class TooOldNodeJSVersion(Exception):
    """Node.js version is too old, raised by
    :meth:`~pycover.pycover.start` or
    :meth:`~pycover.pycover.run`
    """

    def __init__(
        self,
        version_needed: str,
        node_version: str,
    ):
        super().__init__(
            f'Needed node {version_needed}+, '
            'actually installed is '
            f'{node_version}',
        )


class TooOldFipperVersion(Exception):
    """Fipper version is too old, raised by
    :meth:`~pycover.pycover.start` or
    :meth:`~pycover.pycover.run`
    """

    def __init__(
            self,
            version_needed: str,
            fipper_version: str,
    ):
        super().__init__(
            f'Needed fipper {version_needed}+, '
            'actually installed is '
            f'{fipper_version}',
        )


class TooOldTelethonVersion(Exception):
    """Telethon version is too old, raised by
    :meth:`~pycover.pycover.start` or
    :meth:`~pycover.pycover.run`
    """

    def __init__(
        self,
        version_needed: str,
        telethon_version: str,
    ):
        super().__init__(
            f'Needed telethon {version_needed}+, '
            'actually installed is '
            f'{telethon_version}',
        )


class InvalidStreamMode(Exception):
    """The stream mode is invalid, raised by
    :meth:`~pycover.pycover.change_stream` or
    :meth:`~pycover.pycover.join_group_call`
    """

    def __init__(self):
        super().__init__(
            'Invalid stream mode',
        )


class NoMtProtoClientSet(Exception):
    """An MtProto client not set to
    :class:`~pycover.PyCover`, raised by
    :meth:`~pycover.pycover.join_group_call`,
    :meth:`~pycover.pycover.leave_group_call`,
    :meth:`~pycover.pycover.change_volume_call`,
    :meth:`~pycover.pycover.change_stream`,
    :meth:`~pycover.pycover.pause_stream` and
    :meth:`~pycover.pycover.resume_stream`
    """

    def __init__(self):
        super().__init__(
            'No MtProto client set',
        )


class NodeJSNotRunning(Exception):
    """NodeJS core not running, do
    :meth:`~pycover.pycover.start`
    before call these methods, raised by
    :meth:`~pycover.pycover.join_group_call`,
    :meth:`~pycover.pycover.leave_group_call`,
    :meth:`~pycover.pycover.change_volume_call`,
    :meth:`~pycover.pycover.change_stream`,
    :meth:`~pycover.pycover.pause_stream` and
    :meth:`~pycover.pycover.resume_stream`
    """

    def __init__(self):
        super().__init__(
            'Node.js not running',
        )


class NoActiveGroupCall(Exception):
    """No active group call found, raised by
    :meth:`~pycover.pycover.join_group_call`,
    :meth:`~pycover.pycover.leave_group_call`,
    :meth:`~pycover.pycover.change_volume_call`,
    """

    def __init__(self):
        super().__init__(
            'No active group call',
        )


class NotInGroupCallError(Exception):
    """The userbot there isn't in a group call, raised by
    :meth:`~pycover.pycover.leave_group_call`
    """

    def __init__(self):
        super().__init__(
            'The userbot there isn\'t in a group call',
        )


class AlreadyJoinedError(Exception):
    """Already joined into group call, raised by
    :meth:`~pycover.pycover.join_group_call`
    """

    def __init__(self):
        super().__init__(
            'Already joined into group call',
        )


class TelegramServerError(Exception):
    """Telegram Server is having some
    internal problems, raised by
    :meth:`~pycover.pycover.join_group_call`
    """

    def __init__(self):
        super().__init__(
            'Telegram Server is having some '
            'internal problems',
        )


class PyCoverAlreadyRunning(Exception):
    """PyCover client is already running, raised by
    :meth:`~pycover.pycover.start`,
    """

    def __init__(self):
        super().__init__(
            'PyCover client is already running',
        )


class TooManyCustomApiDecorators(Exception):
    """Too Many Custom Api Decorators, raised by
    :meth:`~pycover.CustomApi.on_update_custom_api`,
    """

    def __init__(self):
        super().__init__(
            'Too Many Custom Api Decorators',
        )


class GroupCallNotFound(Exception):
    """Group call not found, raised by
    :meth:`~pycover.pycover.get_active_call`,
    :meth:`~pycover.pycover.get_call`
    """

    def __init__(
        self,
        chat_id: int,
    ):
        super().__init__(
            f'Group call not found with the chat id {chat_id}',
        )


class InvalidMtProtoClient(Exception):
    """You set an invalid MtProto client, raised by
    :meth:`~pycover.PyCover`
    """

    def __init__(self):
        super().__init__(
            'Invalid MtProto Client',
        )


class NoVideoSourceFound(Exception):
    """This error is raised when the stream does not have video streams
    :meth:`~pycover.pycover.join_group_call` or
    :meth:`~pycover.pycover.change_stream`
    """

    def __init__(self, path: str):
        super().__init__(
            f'No video source found on {path}',
        )


class InvalidVideoProportion(Exception):
    """FFmpeg have sent invalid video measure
    response, raised by
    :meth:`~pycover.pycover.join_group_call` or
    :meth:`~pycover.pycover.change_stream`
    """

    def __init__(self, message: str):
        super().__init__(
            message,
        )


class NoAudioSourceFound(Exception):
    """This error is raised when the stream does not have audio streams
    :meth:`~pycover.pycover.join_group_call` or
    :meth:`~pycover.pycover.change_stream`
    """

    def __init__(self, path: str):
        super().__init__(
            f'No audio source found on {path}',
        )


class FFmpegNotInstalled(Exception):
    """FFmpeg isn't installed, this error is raised by
    :meth:`~pycover.pycover.join_group_call` or
    :meth:`~pycover.pycover.change_stream`
    """

    def __init__(self, path: str):
        super().__init__(
            'FFmpeg ins\'t installed on your server',
        )


class RTMPStreamNeeded(Exception):
    """Needed an RTMP Stream, raised by
    :meth:`~pycover.pycover.join_group_call`
    """

    def __init__(self):
        super().__init__(
            'Needed an RTMP Stream',
        )


class UnMuteNeeded(Exception):
    """Needed to unmute the userbot, raised by
    :meth:`~pycover.pycover.join_group_call`
    """

    def __init__(self):
        super().__init__(
            'Needed to unmute the userbot',
        )



from math import gcd

from pycover.exceptions import InvalidVideoProportion
from pycover.types.input_stream.quality import HighQualityVideo
from pycover.types.input_stream.quality import LowQualityVideo
from pycover.types.input_stream.quality import MediumQualityVideo
from pycover.types.input_stream.video_parameters import VideoParameters


def check_support(link: str):
    supported_protocols = [
        'https://',
        'http://',
        'rtmp://',
        'udp://',
    ]
    return any(
        protocol in link
        for protocol in supported_protocols
    )


def check_video_params(
    video_params: VideoParameters,
    dest_width: int,
    dest_height: int,
):
    def resize_ratio(w, h, factor):
        if w > h:
            rescaling = ((1280 if w > 1280 else w) * 100) / w
        else:
            rescaling = ((720 if h > 720 else h) * 100) / h
        h = round((h * rescaling) / 100)
        w = round((w * rescaling) / 100)
        divisor = gcd(w, h)
        ratio_w = w / divisor
        ratio_h = h / divisor
        factor = (divisor * factor) / 100
        return round(ratio_w * factor), round(ratio_h * factor)

    height = video_params.height
    width = video_params.width
    if isinstance(
        video_params,
        HighQualityVideo,
    ):
        width, height = resize_ratio(dest_width, dest_height, 100)
    if isinstance(
        video_params,
        MediumQualityVideo,
    ):
        width, height = resize_ratio(dest_width, dest_height, 66.69)
    if isinstance(
        video_params,
        LowQualityVideo,
    ):
        width, height = resize_ratio(dest_width, dest_height, 50)
    if dest_height < height:
        raise InvalidVideoProportion(
            'Destination height is greater than the original height',
        )
    width = width - 1 if width % 2 else width
    height = height - 1 if height % 2 else height
    return width, height



class StreamType:
    """StreamType, the main means for setting
    the streaming mode.

    Attributes:
        live_stream (``self``):
            Streaming mode for Live Video, or HTTP online video
        local_stream (``self``):
            Streaming mode for Downloaded Video
        pulse_stream (``self``):
            Streaming mode for Live Video, HTTP online video,
            or Local Video
        stream_mode (``int``):
            Get the int value of stream mode
    """

    def __init__(self):
        self._stream_type = 0

    @property
    def live_stream(self):
        self._stream_type = 3
        return self

    @property
    def local_stream(self):
        self._stream_type = 10
        return self

    """
    *** Pulse Stream ***
    Send bytes like a pulsation, and this reduce the slice,
    because the slice is too heavy

    Support: LiveStream, LocalStream
    """
    @property
    def pulse_stream(self):
        self._stream_type = 4
        return self

    @property
    def stream_mode(self):
        return self._stream_type



from sys import platform

from ..types.py_object import PyObject


class ScreenInfo(PyObject):
    def __init__(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        is_primary: bool,
        title: str,
    ):
        self.offset_x = x
        self.offset_y = y
        self.width = width
        self.height = height
        self.is_primary = is_primary
        self.title = title
        self.ffmpeg_parameters = ''

    def build_ffmpeg_command(self, frame_rate: int):
        if platform == 'win32':
            path = 'desktop'
            self.ffmpeg_parameters = '-f gdigrab '
            self.ffmpeg_parameters += f'-offset_x {self.offset_x} '
            self.ffmpeg_parameters += f'-offset_y {self.offset_y} '
        else:
            path = f':0.0+{self.offset_x},{self.offset_y}'
            self.ffmpeg_parameters = '-f x11grab '
        self.ffmpeg_parameters += f'-video_size {self.width}x{self.height} '
        self.ffmpeg_parameters += f'-framerate {frame_rate}'
        return path



from typing import Optional

from ..py_object import PyObject
from .input_audio_stream import InputAudioStream
from .input_video_stream import InputVideoStream


class InputStream(PyObject):
    """The streams descriptor

    Attributes:
        lip_sync (``bool``):
            Lip Sync mode
        stream_audio (:obj:`~pycover.types.InputAudioStream()`):
            Input Audio Stream Descriptor
        stream_video (:obj:`~pycover.types.InputVideoStream()`):
            Input Video Stream Descriptor

    Parameters:
        stream_audio (:obj:`~pycover.types.InputAudioStream()`):
            Audio File Descriptor
        stream_video (:obj:`~pycover.types.InputVideoStream()`):
            Video File Descriptor
        lip_sync (``bool``, **optional**):
            Lip Sync mode
    """

    def __init__(
        self,
        stream_audio: Optional[InputAudioStream] = None,
        stream_video: Optional[InputVideoStream] = None,
        lip_sync: bool = False,
    ):
        self.stream_audio: Optional[InputAudioStream] = stream_audio
        self.stream_video: Optional[InputVideoStream] = stream_video
        self.lip_sync = lip_sync



import asyncio

from ...scaffold import Scaffold
from ...types.call_holder import CallHolder
from ...types.groups import JoinedVoiceChat
from ...types.groups import LeftVoiceChat
from ...types.object import Object
from ...types.stream import ChangedStream
from ...types.stream import PausedStream
from ...types.stream import ResumedStream
from ...types.stream import StreamDeleted


class RawUpdateHandler(Scaffold):
    async def _raw_update_handler(
        self,
        data: dict,
    ):
        obj = Object.from_dict(data)
        solved_update = False
        if 'solver_id' in data:
            solved_update = self._wait_result.resolve_future_update(
                data['solver_id'],
                obj,
            )
        if isinstance(obj, PausedStream):
            self._call_holder.set_status(
                obj.chat_id,
                CallHolder.PAUSED,
            )
        elif isinstance(obj, ResumedStream) or\
                isinstance(obj, ChangedStream) or\
                isinstance(obj, JoinedVoiceChat):
            self._call_holder.set_status(
                obj.chat_id,
                CallHolder.PLAYING,
            )
        elif isinstance(obj, LeftVoiceChat):
            self._call_holder.remove_call(
                obj.chat_id,
            )
        elif isinstance(obj, StreamDeleted):
            self._call_holder.remove_call(
                obj.chat_id,
            )
            asyncio.ensure_future(
                self._binding.send({
                    'action': 'leave_call',
                    'chat_id': obj.chat_id,
                    'type': 'file_deleted',
                }),
            )
        if not solved_update:
            await self._on_event_update.propagate(
                'RAW_UPDATE_HANDLER',
                self,
                obj,
            )
        return {
            'result': 'OK',
        }



import psutil

from ...scaffold import Scaffold


class GetMaxVoiceChat(Scaffold):
    @staticmethod
    def get_max_voice_chat(consumption=5):
        """Get the max number of group calls that can be started

        This method get an estimated number of max group calls

        Parameters:
            consumption (``int``, **optional**):
                Estimated PyCover consumption of Core x Single Group Call

        Returns:
            ``int`` - Max number of group calls can be started

        Example:
            .. code-block:: python
                :emphasize-lines: 5

                from pycover import Client
                from pycover import idle
                ...
                app = Client(client)
                app.get_max_voice_chat()
                app.run()

        """
        core_count = psutil.cpu_count()
        return int((100 / consumption) * core_count)



import asyncio
import logging
import os
from stat import S_ISFIFO
from typing import Dict
from typing import Optional

from aiohttp import ClientConnectorError
from aiohttp import ClientResponse
from aiohttp import ClientSession

from .types.input_stream.video_tools import check_support

py_logger = logging.getLogger('pytgcalls')


class FileManager:
    @staticmethod
    async def check_file_exist(
        path: str,
        headers: Optional[Dict[str, str]] = None,
    ):
        if check_support(path):
            session = ClientSession()
            try:
                response: ClientResponse = await session.get(
                    path,
                    timeout=5,
                    headers=headers,
                )
                response.close()
                if response.status == 200 or \
                        response.status == 403:
                    return
                else:
                    py_logger.info(
                        f'{path} returned with {response.status} code',
                    )
            except ClientConnectorError:
                pass
            except asyncio.exceptions.TimeoutError:
                pass
            finally:
                await session.close()
        if path.startswith('udp://'):
            return
        if S_ISFIFO(os.stat(path).st_mode):
            return
        if not os.path.isfile(path):
            raise FileNotFoundError()



import asyncio
from typing import Union

from ...exceptions import NoActiveGroupCall
from ...exceptions import NodeJSNotRunning
from ...exceptions import NoMtProtoClientSet
from ...exceptions import NotInGroupCallError
from ...mtproto import BridgedClient
from ...scaffold import Scaffold
from ...types import NotInGroupCall
from ...types.session import Session


class LeaveGroupCall(Scaffold):
    async def leave_group_call(
        self,
        chat_id: Union[int, str],
    ):
        """Leave a group call

        This method allow to leave a Group Call

        Parameters:
             chat_id (``int`` | ``str``):
                Unique identifier of the target chat.
                Can be a direct id (int) or a username (str)

        Raises:
            NoMtProtoClientSet: In case you try
                to call this method without any MtProto client
            NodeJSNotRunning: In case you try
                to call this method without do
                :meth:`~pycover.pycover.start` before
            NoActiveGroupCall: In case you try
                to edit a not started group call
            NotInGroupCallError: In case you try
                to leave a non-joined group call

        Example:
            .. code-block:: python
                :emphasize-lines: 10-12

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client)
                app.start()

                ...  # Call API methods

                app.leave_group_call(
                    -1001185324811,
                )

                idle()
        """
        if self._app is not None:
            if self._wait_until_run is not None:
                try:
                    chat_id = int(chat_id)
                except ValueError:
                    chat_id = BridgedClient.chat_id(
                        await self._app.resolve_peer(chat_id),
                    )
                chat_call = await self._app.get_full_chat(
                    chat_id,
                )
                if chat_call is not None:
                    solver_id = Session.generate_session_id(24)

                    async def internal_sender():
                        if not self._wait_until_run.done():
                            await self._wait_until_run
                        await self._binding.send({
                            'action': 'leave_call',
                            'chat_id': chat_id,
                            'type': 'requested',
                            'solver_id': solver_id,
                        })
                    asyncio.ensure_future(internal_sender())
                    result = await self._wait_result.wait_future_update(
                        solver_id,
                    )
                    if isinstance(result, NotInGroupCall):
                        raise NotInGroupCallError()
                else:
                    raise NoActiveGroupCall()
            else:
                raise NodeJSNotRunning()
        else:
            raise NoMtProtoClientSet()



from asyncio.log import logger

from pycover.scaffold import Scaffold


class SetVideoCallStatus(Scaffold):
    async def _set_video_call_status(
        self,
        params: dict,
    ):
        result = {
            'result': 'OK',
        }
        stopped_status = params['stopped_status'] \
            if 'stopped_status' in params else None
        paused_status = params['paused_status'] \
            if 'paused_status' in params else None
        try:
            chat_id = int(params['chat_id'])
            await self._app.set_video_call_status(
                chat_id,
                stopped_status,
                paused_status,
                self._cache_user_peer.get(chat_id),
            )
        except Exception as e:
            logger.error(f'SetVideoCallStatus: {e}')
            result = {
                'result': str(e),
            }
        return result



from ..video_parameters import VideoParameters


class HighQualityVideo(VideoParameters):
    """High Video Quality (1280x720)

    Attributes:
        width (``int``):
            Video width
        height (``int``):
            Video height
        frame_rate (``int``):
            Framerate of video
    """

    def __init__(self):
        super().__init__(
            1280,
            720,
            20,
        )



import asyncio
import logging
import shlex
from typing import Union

from ...exceptions import AlreadyJoinedError
from ...exceptions import InvalidStreamMode
from ...exceptions import NoActiveGroupCall
from ...exceptions import NodeJSNotRunning
from ...exceptions import NoMtProtoClientSet
from ...exceptions import RTMPStreamNeeded
from ...exceptions import TelegramServerError
from ...exceptions import UnMuteNeeded
from ...file_manager import FileManager
from ...mtproto import BridgedClient
from ...scaffold import Scaffold
from ...stream_type import StreamType
from ...types import AlreadyJoined
from ...types import CaptureAudioDevice
from ...types import CaptureAVDesktop
from ...types import CaptureAVDeviceDesktop
from ...types import CaptureVideoDesktop
from ...types import ErrorDuringJoin
from ...types import MutedCall
from ...types import UpgradeNeeded
from ...types.input_stream import AudioPiped
from ...types.input_stream import AudioVideoPiped
from ...types.input_stream import InputStream
from ...types.input_stream import VideoPiped
from ...types.input_stream.audio_image_piped import AudioImagePiped
from ...types.session import Session

py_logger = logging.getLogger('pytgcalls')


class JoinGroupCall(Scaffold):
    async def join_group_call(
        self,
        chat_id: Union[int, str],
        stream: InputStream,
        invite_hash: str = None,
        join_as=None,
        stream_type: StreamType = None,
    ):
        """Join a group call to stream a file

        This method allow to stream a file to Telegram
        Group Calls

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier of the target chat.
                Can be a direct id (int) or a username (str)
            stream (:obj:`~pycover.types.InputStream()`):
                Input Streams descriptor, can be used also
                :obj:`~pycover.types.AudioPiped()`,
                :obj:`~pycover.types.AudioImagePiped()`,
                :obj:`~pycover.types.AudioVideoPiped()` or
                :obj:`~pycover.types.VideoPiped()`
            invite_hash (``str``, **optional**):
                Unique identifier for the invite in a group call
                in form of a t.me link
            join_as (`InputPeer (P)`_ | `InputPeer (T)`_, **optional**):
                InputPeer of join as channel or profile
            stream_type (:obj:`~pycover.StreamType`, **optional**)
                The type of Stream

        Raises:
            NoMtProtoClientSet: In case you try
                to call this method without any MtProto client
            NodeJSNotRunning: In case you try
                to call this method without do
                :meth:`~pycover.pycover.start` before
            NoActiveGroupCall: In case you try
                to edit a not started group call
            FileNotFoundError: In case you try
                a non-existent file
            InvalidStreamMode: In case you try
                to set a void stream mode
            FFmpegNotInstalled: In case you try
                to use the Piped input stream, and
                you don't have ffmpeg installed
            NoAudioSourceFound: In case you try
                to play an audio file from a file
                without the sound
            NoVideoSourceFound: In case you try
                to play a video file from a file
                without the video
            InvalidVideoProportion: In case you try
                to play a video without correct
                proportions
            AlreadyJoinedError: In case you try
                to join in already joined group
                call
            TelegramServerError: Error occurred when
                joining to a group call (
                Telegram Server Side)
            RTMPStreamNeeded: In case you try
                to join a group call without
                a RTMP stream
            UnMuteNeeded: In case you try
                to play on a muted group call

        Example:
            .. code-block:: python
                :emphasize-lines: 10-15

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client)
                app.start()

                ...  # Call API methods

                app.join_group_call(
                    -1001185324811,
                    AudioPiped(
                        'test.mp4',
                    )
                )

                idle()
        """
        if join_as is None:
            join_as = self._cache_local_peer
        if stream_type is None:
            stream_type = StreamType().local_stream
        if stream_type.stream_mode == 0:
            raise InvalidStreamMode()
        try:
            chat_id = int(chat_id)
        except ValueError:
            chat_id = BridgedClient.chat_id(
                await self._app.resolve_peer(chat_id),
            )
        self._cache_user_peer.put(chat_id, join_as)
        headers = None
        if isinstance(
            stream,
            AudioImagePiped,
        ) or isinstance(
            stream,
            AudioPiped,
        ) or isinstance(
            stream,
            AudioVideoPiped,
        ) or isinstance(
            stream,
            VideoPiped,
        ):
            headers = stream.raw_headers
        if stream.stream_video is not None:
            if not stream.stream_video.path.startswith('screen://'):
                await FileManager.check_file_exist(
                    stream.stream_video.path.replace(
                        'fifo://',
                        '',
                    ).replace(
                        'image:',
                        '',
                    ),
                    headers,
                )
        if stream.stream_audio is not None:
            if not stream.stream_audio.path.startswith('device://'):
                await FileManager.check_file_exist(
                    stream.stream_audio.path.replace(
                        'fifo://',
                        '',
                    ).replace(
                        'image:',
                        '',
                    ),
                    headers,
                )
        audio_f_parameters = ''
        video_f_parameters = ''
        if isinstance(
            stream,
            AudioImagePiped,
        ) or isinstance(
            stream,
            AudioPiped,
        ) or isinstance(
            stream,
            AudioVideoPiped,
        ) or isinstance(
            stream,
            VideoPiped,
        ) or isinstance(
            stream,
            CaptureVideoDesktop,
        ) or isinstance(
            stream,
            CaptureAudioDevice,
        ):
            await stream.check_pipe()
            if stream.stream_audio:
                if stream.stream_audio.header_enabled:
                    audio_f_parameters = stream.headers
            audio_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.ffmpeg_parameters),
            )
            if stream.stream_video:
                if stream.stream_video.header_enabled:
                    video_f_parameters = stream.headers
            video_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.ffmpeg_parameters),
            )
        elif isinstance(
            stream,
            CaptureAVDeviceDesktop,
        ):
            audio_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.audio_ffmpeg),
            )
            video_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.video_ffmpeg),
            )
        elif isinstance(
            stream,
            CaptureAVDesktop,
        ):
            await stream.check_pipe()
            if stream.stream_audio:
                if stream.stream_audio.header_enabled:
                    audio_f_parameters = stream.headers
            audio_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.audio_ffmpeg),
            )
            video_f_parameters += ':_cmd_:'.join(
                shlex.split(stream.video_ffmpeg),
            )
        if self._app is not None:
            if self._wait_until_run is not None:
                if not self._wait_until_run.done():
                    await self._wait_until_run
                chat_call = await self._app.get_full_chat(
                    chat_id,
                )
                stream_audio = stream.stream_audio
                stream_video = stream.stream_video
                if chat_call is not None:
                    solver_id = Session.generate_session_id(24)

                    async def internal_sender():
                        request = {
                            'action': 'join_call',
                            'chat_id': chat_id,
                            'invite_hash': invite_hash,
                            'buffer_long': stream_type.stream_mode,
                            'lip_sync': stream.lip_sync,
                            'solver_id': solver_id,
                        }
                        if stream_audio is not None:
                            request['stream_audio'] = {
                                'path': stream_audio.path,
                                'bitrate': stream_audio.parameters.bitrate,
                                'ffmpeg_parameters': audio_f_parameters,
                            }
                        if stream_video is not None:
                            video_parameters = stream_video.parameters
                            if video_parameters.frame_rate % 5 != 0 and \
                                    not isinstance(stream, AudioImagePiped):
                                py_logger.warning(
                                    'For better experience the '
                                    'video frame rate must be a multiple of 5',
                                )
                            request['stream_video'] = {
                                'path': stream_video.path,
                                'width': video_parameters.width,
                                'height': video_parameters.height,
                                'framerate': video_parameters.frame_rate,
                                'ffmpeg_parameters': video_f_parameters,
                            }
                        await self._binding.send(request)
                    asyncio.ensure_future(internal_sender())
                    result = await self._wait_result.wait_future_update(
                        solver_id,
                    )
                    if isinstance(result, AlreadyJoined):
                        raise AlreadyJoinedError()
                    elif isinstance(result, ErrorDuringJoin):
                        raise TelegramServerError()
                    elif isinstance(result, UpgradeNeeded):
                        raise RTMPStreamNeeded()
                    elif isinstance(result, MutedCall):
                        raise UnMuteNeeded()
                else:
                    raise NoActiveGroupCall()
            else:
                raise NodeJSNotRunning()
        else:
            raise NoMtProtoClientSet()



from pycover.types.py_object import PyObject


class Update(PyObject):
    """A Raw Update

    Attributes:
        chat_id (``int``):
            Unique identifier of chat.

    Parameters:
        chat_id (``int``):
            Unique identifier of chat.

    """

    def __init__(
        self,
        chat_id: int,
    ):
        self.chat_id = chat_id



from typing import Dict
from typing import Optional

from ...ffprobe import FFprobe
from .audio_parameters import AudioParameters
from .input_audio_stream import InputAudioStream
from .input_stream import InputStream
from .input_video_stream import InputVideoStream
from .video_parameters import VideoParameters
from .video_tools import check_video_params


class AudioVideoPiped(InputStream):
    """The audio/video stream piped descriptor

    Attributes:
        ffmpeg_parameters (``str``):
            FFmpeg additional parameters
        lip_sync (``bool``):
            Lip Sync mode
        raw_headers (``str``):
            Headers of http the connection
        stream_audio (:obj:`~pycover.types.InputAudioStream()`):
            Input Audio Stream Descriptor
        stream_video (:obj:`~pycover.types.InputVideoStream()`):
            Input Video Stream Descriptor

    Parameters:
        path (``str``):
            The audio-video (Like MP4 or etc) file path
        audio_parameters (:obj:`~pycover.types.AudioParameters()`):
            The audio parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityAudio()`,
            :obj:`~pycover.types.MediumQualityAudio()` or
            :obj:`~pycover.types.LowQualityAudio()`
        video_parameters (:obj:`~pycover.types.VideoParameters()`):
            The video parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityVideo()`,
            :obj:`~pycover.types.MediumQualityVideo()` or
            :obj:`~pycover.types.LowQualityVideo()`
        headers (``Dict[str, str]``, **optional**):
            Headers of http the connection
        additional_ffmpeg_parameters (``str``, **optional**):
            FFmpeg additional parameters
    """

    def __init__(
        self,
        path: str,
        audio_parameters: AudioParameters = AudioParameters(),
        video_parameters: VideoParameters = VideoParameters(),
        headers: Optional[Dict[str, str]] = None,
        additional_ffmpeg_parameters: str = '',
    ):
        self._path = path
        self.ffmpeg_parameters = additional_ffmpeg_parameters
        self.raw_headers = headers
        super().__init__(
            InputAudioStream(
                f'fifo://{path}',
                audio_parameters,
            ),
            InputVideoStream(
                f'fifo://{path}',
                video_parameters,
            ),
        )
        self.lip_sync = True

    @property
    def headers(self):
        return FFprobe.ffmpeg_headers(self.raw_headers)

    async def check_pipe(self):
        dest_width, dest_height, header = await FFprobe.check_file(
            self._path,
            needed_audio=True,
            needed_video=True,
            needed_image=False,
            headers=self.raw_headers,
        )
        width, height = check_video_params(
            self.stream_video.parameters,
            dest_width,
            dest_height,
        )
        self.stream_video.header_enabled = header
        self.stream_audio.header_enabled = header
        self.stream_video.parameters.width = width
        self.stream_video.parameters.height = height



from .core import Core
from .decorators import Decorators
from .groups import Groups
from .handlers import Handlers
from .stream import Stream
from .utilities import Utilities


class Methods(
    Core,
    Decorators,
    Groups,
    Handlers,
    Stream,
    Utilities,
):
    pass



from pycover.types.py_object import PyObject


class GroupCallParticipant(PyObject):
    """Info about a group call participant

    Attributes:
        user_id (``int``):
            Unique identifier of participant.
        muted (``bool``):
            Whether the participant is muted
        muted_by_admin (``bool``):
            Whether the participant is muted by
            admin
        video (``bool``):
            Whether this participant is currently
            broadcasting a video stream
        screen_sharing (``bool``):
            Whether this participant is currently
            broadcasting a screen sharing
        video_camera (``bool``):
            Whether this participant is currently
            broadcasting a video camera
        raised_hand (``bool``):
            Whether this participant have
            raised the hand
        volume (``int``):
            Volume level of the participant

    Parameters:
        user_id (``int``):
            Unique identifier of participant.
        muted (``bool``):
            Telegram API parameter.
        muted_by_admin (``bool``):
            Telegram API parameter.
        video (``bool``):
            Telegram API parameter.
        screen_sharing (``bool``):
            Telegram API parameter.
        video_camera (``bool``):
            Telegram API parameter.
        raised_hand (``bool``):
            Telegram API parameter.
        volume (``int``):
            Telegram API parameter.
    """

    def __init__(
        self,
        user_id: int,
        muted: bool,
        muted_by_admin: bool,
        video: bool,
        screen_sharing: bool,
        video_camera: bool,
        raised_hand: bool,
        volume: int,
    ):
        self.user_id: int = user_id
        self.muted: bool = muted
        self.muted_by_admin: bool = muted_by_admin
        self.video: bool = video
        self.screen_sharing: bool = screen_sharing
        self.video_camera: bool = video_camera
        self.raised_hand: bool = raised_hand
        self.volume: int = volume



from .cache_peer import CachePeer
from .get_max_voice_chat import GetMaxVoiceChat
from .is_connected import IsConnected
from .mtproto_handler import MtProtoHandler
from .ping import Ping
from .run import Run
from .start import Start


class Utilities(
    CachePeer,
    GetMaxVoiceChat,
    IsConnected,
    Ping,
    MtProtoHandler,
    Run,
    Start,
):
    pass



from ...types.groups.group_call_participant import GroupCallParticipant
from ...types.update import Update


class UpdatedGroupCallParticipant(Update):
    """A participant have changed
    him status

    Attributes:
        chat_id (``int``):
            Unique identifier of chat.
        participant (:obj:`~pycover.types.GroupCallParticipant()`):
            Info about a group call participant

    Parameters:
        chat_id (``int``):
            Unique identifier of chat.
        participant (:obj:`~pycover.types.GroupCallParticipant()`):
            Info about a group call participant
    """

    def __init__(
        self,
        chat_id: int,
        participant: GroupCallParticipant,
    ):
        super().__init__(chat_id)
        self.participant = participant



from ...scaffold import Scaffold
from .idle import idle


class Run(Scaffold):
    async def run(self):
        """Start the client, idle the main script and finally stop the client.

        This is a convenience method that calls start(), idle() and stop
        in sequence. It makes running a client less verbose, but is not
        suitable in case you want to run more than one client in a single
        main script, since idle() will block after starting the own client.

        Raises:
            PyCoverAlreadyRunning: In case you try
                to start an already started client.

        Example:
            .. code-block:: python
                :emphasize-lines: 8

                from pycover import Client
                from pycover import idle
                ...
                app = Client(client)

                ...  # Call API decorators / MtProto decorators

                app.run()
        """
        await self.start()
        await idle()



from typing import List
from typing import Optional
from typing import Union

from ...mtproto import BridgedClient
from ...scaffold import Scaffold
from ...types.groups.group_call_participant import GroupCallParticipant


class GetParticipants(Scaffold):
    async def get_participants(
        self,
        chat_id: Union[int, str],
    ) -> Optional[List[GroupCallParticipant]]:
        """Get list of participants from a group call

        This method return the list of participants on a group call
        using MtProto APIs

        Parameters:
            chat_id (``int`` | ``str``):
                Can be a direct id (int) or a username (str)

        Returns:
            List of :obj:`~pycover.types.GroupCallParticipant()`:
            On success, a list of participants is returned

        Example:
            .. code-block:: python
                :emphasize-lines: 10-12

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client)
                app.start()

                ...  # Call API methods

                app.get_participants(
                    -1001185324811,
                )

                idle()
        """
        try:
            chat_id = int(chat_id)
        except ValueError:
            chat_id = BridgedClient.chat_id(
                await self._app.resolve_peer(chat_id),
            )
        self._call_holder.get_call(
            chat_id,
        )
        return await self._app.get_group_call_participants(
            chat_id,
        )



import asyncio
import logging
import signal
from signal import SIGABRT
from signal import SIGINT
from signal import signal as signal_fn
from signal import SIGTERM


py_logger = logging.getLogger('pytgcalls')

is_idling = False

# Signal number to name
signals = {
    k: v for v, k in signal.__dict__.items()
    if v.startswith('SIG') and not v.startswith('SIG_')
}


async def idle():
    """Block the main script execution until a signal is received.

    This function will run indefinitely in order to block the main
    script execution and prevent it from exiting while having client(s)
    that are still running in the background.

    The way PyCover works, it will keep your handlers in a pool of
    worker threads, which are executed concurrently outside the main
    thread; calling idle() will ensure the client(s) will be kept alive
    by not letting the main script to end, until you decide to quit.

    Once a signal is received (e.g.: from CTRL+C) the function will
    terminate and your main script will continue.

    Example:
        .. code-block:: python
            :emphasize-lines: 15

            from pycover import Client
            from pycover import idle
            ...

            app1 = PyCover(client1)
            app2 = PyCover(client2)
            app3 = PyCover(client3)

            ...  # Set handlers up

            app1.start()
            app2.start()
            app3.start()

            idle()
    """
    global is_idling

    def signal_handler(signum, __):
        global is_idling
        py_logger.info(f'Stop signal received ({signals[signum]}). Exiting...')
        is_idling = False
    for s in (SIGINT, SIGTERM, SIGABRT):
        signal_fn(s, signal_handler)
    is_idling = True
    while is_idling:
        await asyncio.sleep(1)



from typing import Callable

from ...scaffold import Scaffold


class OnGroupCallInvite(Scaffold):
    def on_group_call_invite(self) -> Callable:
        """Decorator for handling when invited
        from voice chat event.

        When your userbot will be invited on voice
        chat, this decorator will be raised

        Example:
            .. code-block:: python
                :emphasize-lines: 4-5

                ...
                app = PyCover(client)
                ...
                @app.on_group_call_invite()
                async def handler(client: PyCover, service_msg):
                    print(service_msg)
                ...
                app.run()

        """
        method = 'INVITE_HANDLER'

        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._on_event_update.add_handler(
                    method,
                    func,
                )
            return func
        return decorator



from typing import Union

from ...exceptions import NoActiveGroupCall
from ...exceptions import NodeJSNotRunning
from ...exceptions import NoMtProtoClientSet
from ...mtproto import BridgedClient
from ...scaffold import Scaffold


class ChangeVolumeCall(Scaffold):
    async def change_volume_call(
        self,
        chat_id: Union[int, str],
        volume: int,
    ):
        """Change the volume of the playing stream

        This method change the output volume of the userbot
        using MtProto APIs

        Parameters:
            chat_id (``int`` | ``str``):
                Can be a direct id (int) or a username (str)
            volume (``int``)
                Volume to set to the stream

        Raises:
            NoMtProtoClientSet: In case you try
                to call this method without any MtProto client
            NodeJSNotRunning: In case you try
                to call this method without do
                :meth:`~pycover.pycover.start` before
            NoActiveGroupCall: In case you try
                to edit a not started group call

        Example:
            .. code-block:: python
                :emphasize-lines: 10-13

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client)
                app.start()

                ...  # Call API methods

                app.change_volume_call(
                    -1001185324811,
                    175,
                )

                idle()
        """
        if self._app is not None:
            if self._wait_until_run is not None:
                try:
                    chat_id = int(chat_id)
                except ValueError:
                    chat_id = BridgedClient.chat_id(
                        await self._app.resolve_peer(chat_id),
                    )
                if not self._wait_until_run.done():
                    await self._wait_until_run
                chat_call = await self._app.get_full_chat(
                    chat_id,
                )
                if chat_call is not None:
                    await self._app.change_volume(
                        chat_id,
                        volume,
                        self._cache_user_peer.get(chat_id),
                    )
                else:
                    raise NoActiveGroupCall()
            else:
                raise NodeJSNotRunning()
        else:
            raise NoMtProtoClientSet()



from typing import Callable

from ...scaffold import Scaffold


class OnClosedVoiceChat(Scaffold):
    def on_closed_voice_chat(self) -> Callable:
        """Decorator for handling closed voice chat event.

        When a video chat closes, this decorator will
        be raised

        Example:
            .. code-block:: python
                :emphasize-lines: 4-5

                ...
                app = PyCover(client)
                ...
                @app.on_closed_voice_chat()
                async def handler(client: PyCover, chat_id: int):
                    print(chat_id)
                ...
                app.run()

        """
        method = 'CLOSED_HANDLER'

        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._on_event_update.add_handler(
                    method,
                    func,
                )
            return func
        return decorator



from ...scaffold import Scaffold


class LeaveVoiceCall(Scaffold):
    async def _leave_voice_call(
        self,
        params: dict,
    ):
        result = {
            'result': 'OK',
        }
        try:
            await self._app.leave_group_call(
                int(params['chat_id']),
            )
        except Exception as e:
            result = {
                'result': str(e),
            }
        return result



from json import JSONDecodeError
from typing import Callable
from typing import Optional

from aiohttp import web
from aiohttp.web_request import BaseRequest

from ..exceptions import TooManyCustomApiDecorators


class CustomApi:
    """CustomAPI, the main means host the local http api server.

    Parameters:
        port (``int``, **optional**):
            CustomApi port to bind the API http server
    """

    def __init__(
        self,
        port: int = 24859,
    ):
        self._handler: Optional[Callable] = None
        self._app: web.Application = web.Application()
        self._runner: Optional[web.AppRunner] = None
        self._port = port

    def on_update_custom_api(self) -> Callable:
        """Decorator for handling when received http call
        to api backend

        Raises:
            TooManyCustomApiDecorators: In case you try
                to add to much decorators to a single
                CustomApi instance

        Example:
            .. code-block:: python
                :emphasize-lines: 4-5

                ...
                webserver = CustomApi(client)
                ...
                @webserver.on_update_custom_api()
                async def handler(request: dict):
                    print(update)
                    ... # Add Your Code here, switch or etc
                    return {
                        'result': 'OK',
                    }
                webserver.start()
                idle()

        """

        if self._handler is None:
            def decorator(func: Callable) -> Callable:
                if self is not None:
                    self._handler = func
                return func

            return decorator
        else:
            raise TooManyCustomApiDecorators()

    async def start(self):
        """Start the Custom Api.

        This method start the internal http webserver,
        this is helpful if you need an API interface to pycover.

        Example:
            .. code-block:: python
                :emphasize-lines: 5

                from pycover import CustomApi
                from pycover import idle

                webserver = CustomApi()
                webserver.start()

                idle()
        """
        async def on_update(request: BaseRequest):
            try:
                params = await request.json()
            except JSONDecodeError:
                return web.json_response({
                    'result': 'INVALID_JSON_FORMAT_REQUEST',
                })
            if self._handler is not None:
                result = await self._handler(params)
                if isinstance(result, dict) or \
                        isinstance(result, list):
                    return web.json_response(result)
                else:
                    return web.json_response({
                        'result': 'INVALID_RESPONSE',
                    })
            else:
                return web.json_response({
                    'result': 'NO_CUSTOM_API_DECORATOR',
                })

        self._app.router.add_post(
            '/',
            on_update,
        )
        runner = web.AppRunner(self._app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', self._port)
        await site.start()



from ..video_parameters import VideoParameters


class LowQualityVideo(VideoParameters):
    """Low Video Quality (640x360)

    Attributes:
        width (``int``):
            Video width
        height (``int``):
            Video height
        frame_rate (``int``):
            Framerate of video
    """

    def __init__(self):
        super().__init__(
            640,
            360,
            20,
        )



from typing import Union

from ...mtproto import BridgedClient
from ...scaffold import Scaffold


class GetCall(Scaffold):
    async def get_call(
        self,
        chat_id: Union[int, str],
    ):
        """Check/Get an existent group call

        This method check if exist a Group Call,
        if not, this can raise an error

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier of the target chat.
                Can be a direct id (int) or a username (str)

        Raises:
            GroupCallNotFound: In case you try
                to get a non-existent group call

        Returns:
            :obj:`~pycover.types.GroupCall()`: On success,
            the group call is returned.

        Example:
            .. code-block:: python
                :emphasize-lines: 10-12

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client1)
                app.start()

                ...  # Call API methods

                app.get_call(
                    -1001185324811,
                )

                idle()
        """
        try:
            chat_id = int(chat_id)
        except ValueError:
            chat_id = BridgedClient.chat_id(
                await self._app.resolve_peer(chat_id),
            )
        return self._call_holder.get_call(
            chat_id,
        )



import asyncio

from ...exceptions import PyCoverAlreadyRunning
from ...scaffold import Scaffold


class Start(Scaffold):
    async def start(self):
        """Start the client.

        This method start and then connects to the NodeJS core.

        Raises:
            PyCoverAlreadyRunning: In case you try
                to start an already started client.

        Example:
            .. code-block:: python
                :emphasize-lines: 5

                from pycover import Client
                from pycover import idle
                ...
                app = Client(client)
                app.start()

                ...  # Call API methods

                idle()
        """
        if not self._is_running:
            self._is_running = True
            loop = asyncio.get_running_loop()
            self._wait_until_run = loop.create_future()
            self._env_checker.check_environment()
            await self._init_mtproto()
            self._handle_mtproto()
            await self._start_binding()
        else:
            raise PyCoverAlreadyRunning()



from .groups import AlreadyJoined
from .groups import ErrorDuringJoin
from .groups import JoinedVoiceChat
from .groups import LeftVoiceChat
from .groups import MutedCall
from .groups import NotInGroupCall
from .groups import UpgradeNeeded
from .stream import ChangedStream
from .stream import MutedStream
from .stream import PausedStream
from .stream import ResumedStream
from .stream import StreamDeleted
from .stream import StreamTime
from .stream import UnMutedStream


class Object:
    @staticmethod
    def from_dict(
        data: dict,
    ):
        event_name = data['result']
        chat_id = int(data['chat_id'])
        if event_name == 'PAUSED_STREAM':
            return PausedStream(chat_id)
        elif event_name == 'RESUMED_STREAM':
            return ResumedStream(chat_id)
        elif event_name == 'CHANGED_STREAM':
            return ChangedStream(chat_id)
        elif event_name == 'JOINED_VOICE_CHAT':
            return JoinedVoiceChat(chat_id)
        elif event_name == 'NOT_IN_GROUP_CALL':
            return NotInGroupCall(chat_id)
        elif event_name == 'JOIN_ERROR':
            return ErrorDuringJoin(chat_id)
        elif event_name == 'ALREADY_JOINED':
            return AlreadyJoined(chat_id)
        elif event_name == 'LEFT_VOICE_CHAT':
            return LeftVoiceChat(chat_id)
        elif event_name == 'STREAM_DELETED':
            return StreamDeleted(chat_id)
        elif event_name == 'MUTED_STREAM':
            return MutedStream(chat_id)
        elif event_name == 'UNMUTED_STREAM':
            return UnMutedStream(chat_id)
        elif event_name == 'APP_UPGRADE_NEEDED':
            return UpgradeNeeded(chat_id)
        elif event_name == 'PLAYED_TIME':
            return StreamTime(data['time'])
        elif event_name == 'UNMUTE_NEEDED':
            return MutedCall(chat_id)



from ..py_object import PyObject
from .video_parameters import VideoParameters


class InputVideoStream(PyObject):
    """The raw video stream (RAW_VIDEO) descriptor

    Attributes:
        path (``str``):
            The video file path
        parameters (:obj:`~pycover.types.VideoParameters()`):
            The video parameters of the stream

    Parameters:
        path (``str``):
            The video file path (RAW_VIDEO format only)
        parameters (:obj:`~pycover.types.VideoParameters()`):
            The video parameters of the stream, can be used also
            :obj:`~pycover.types.HighQualityVideo()`,
            :obj:`~pycover.types.MediumQualityVideo()` or
            :obj:`~pycover.types.LowQualityVideo()`
    """

    def __init__(
        self,
        path: str,
        parameters: VideoParameters = VideoParameters(),
        header_enabled: bool = False,
    ):
        self.path: str = path
        self.parameters: VideoParameters = parameters
        self.header_enabled: bool = header_enabled



from ..py_object import PyObject


class AudioParameters(PyObject):
    """Audio parameters of the stream

    Attributes:
        bitrate (``int``):
            Audio bitrate

    Parameters:
        bitrate (``int``):
            Audio bitrate (0 - 48000 max
            allowed by Telegram)
    """

    def __init__(
        self,
        bitrate: int = 48000,
    ):
        self.bitrate: int = 48000 if bitrate > 48000 else bitrate



from .binding_runner import BindingRunner
from .join_voice_call import JoinVoiceCall
from .leave_voice_call import LeaveVoiceCall
from .set_video_call_status import SetVideoCallStatus


class Core(
    BindingRunner,
    JoinVoiceCall,
    LeaveVoiceCall,
    SetVideoCallStatus,
):
    pass



from typing import Dict

from ..exceptions import GroupCallNotFound
from .groups import GroupCall
from .list import List


class CallHolder:
    PLAYING = 1
    PAUSED = 2
    IDLE = 3

    def __init__(self):
        self._calls: Dict[int, int] = {}

    def set_status(
        self,
        chat_id: int,
        status: int,
    ):
        self._calls[chat_id] = status

    @property
    def active_calls(self):
        return List([
            GroupCall(x, self._calls[x]) for x in self._calls
            if self._calls[x] != self.IDLE
        ])

    @property
    def calls(self):
        return List([
            GroupCall(x, self._calls[x]) for x in self._calls
        ])

    def get_active_call(
        self,
        chat_id: int,
    ):
        if chat_id in self._calls:
            if self._calls[chat_id] != self.IDLE:
                return GroupCall(chat_id, self._calls[chat_id])
        raise GroupCallNotFound(chat_id)

    def get_call(
        self,
        chat_id: int,
    ):
        if chat_id in self._calls:
            return GroupCall(chat_id, self._calls[chat_id])
        raise GroupCallNotFound(chat_id)

    def remove_call(
        self,
        chat_id: int,
    ):
        if chat_id in self._calls:
            del self._calls[chat_id]



import logging

from ...scaffold import Scaffold

py_logger = logging.getLogger('pytgcalls')


class JoinVoiceCall(Scaffold):
    async def _join_voice_call(
        self,
        params: dict,
    ):
        request_call = {
            'ufrag': params['ufrag'],
            'pwd': params['pwd'],
            'fingerprints': [{
                'hash': params['hash'],
                'setup': params['setup'],
                'fingerprint': params['fingerprint'],
            }],
            'ssrc': params['source'],
        }
        if params['source_groups'] is not None:
            request_call['ssrc-groups'] = [
                {
                    'semantics': 'FID',
                    'sources': [ssrc for ssrc in params['source_groups']],
                },
            ]
        chat_id = int(params['chat_id'])
        try:
            return await self._app.join_group_call(
                chat_id,
                request_call,
                params['invite_hash'],
                params['have_video'],
                self._cache_user_peer.get(chat_id),
            )
        except Exception as e:
            if 'GROUPCALL_FORBIDDEN' in str(e):
                self._cache_user_peer.pop(chat_id)
            py_logger.error(f'JOIN_VOICE_CALL_ERROR -> {e}')
            return {'transport': None, 'error': str(e)}



import os
import re
import shutil
import subprocess
import sys

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_ext import build_ext

base_path = os.path.abspath(os.path.dirname(__file__))


class NodeJsExtension(Extension):
    def __init__(self, name, source_dir=''):
        super().__init__(name, sources=[])
        self.source_dir = os.path.abspath(source_dir)


class SetupHelper:
    def __init__(
        self,
        source_dir: str,
        ext_dir: str,
        tmp_dir: str,
    ):
        folder_package = ''
        for item in sys.path:
            if 'dist-packages' in item or 'site-packages' in item:
                folder_package = item
                break
        self._source_dir = source_dir
        self._ext_dir = ext_dir
        self._tmp_dir = tmp_dir
        self._folder_package = folder_package

    def clean_old_installation(self):
        # CHECK IF IS INSTALLATION FROM PYPI
        if os.path.isdir(os.path.join(self._source_dir, 'src')):
            try:
                shutil.rmtree(
                    os.path.join(
                        self._folder_package,
                        'pycover', 'node_modules',
                    ),
                )
            except OSError:
                pass
            try:
                shutil.rmtree(
                    os.path.join(
                        self._folder_package, 'pycover', 'dist',
                    ),
                ),
            except OSError:
                pass
            try:
                shutil.rmtree(os.path.join(self._tmp_dir))
            except OSError:
                pass

    def run_installation(self):
        # CHECK IF IS INSTALLATION FROM PYPI
        if os.path.isdir(os.path.join(self._source_dir, 'src')):
            # COPY NEEDED FILES
            shutil.copytree(
                os.path.join(self._source_dir, 'src'),
                os.path.join(self._tmp_dir, 'src'),
            )
            shutil.copyfile(
                os.path.join(self._source_dir, 'package.json'),
                os.path.join(self._tmp_dir, 'package.json'),
            )
            shutil.copyfile(
                os.path.join(self._source_dir, 'tsconfig.json'),
                os.path.join(self._tmp_dir, 'tsconfig.json'),
            )
            shutil.copyfile(
                os.path.join(self._source_dir, '.npmignore'),
                os.path.join(self._tmp_dir, '.npmignore'),
            )
            # START COMPILATION
            subprocess.check_call(
                'npm install .',
                shell=True,
                cwd=self._tmp_dir,
            )
            shutil.copytree(
                os.path.join(self._tmp_dir, 'node_modules'),
                os.path.join(self._ext_dir, 'pycover', 'node_modules'),
            )
            shutil.copytree(
                os.path.join(self._tmp_dir, 'dist'),
                os.path.join(self._ext_dir, 'pycover', 'dist'),
            )


class NodeJsBuilder(build_ext):
    def run(self):
        super().run()

    def build_extension(self, ext):
        ext_dir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)),
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        sh = SetupHelper(
            ext.source_dir,
            ext_dir,
            self.build_temp,
        )
        sh.clean_old_installation()
        sh.run_installation()


with open(os.path.join(base_path, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

setup(
    name='pycover',
    version='0.9.8',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/AyiinXd/pycover',
    author='AyiinXd',
    author_email='buayadanau4@gmail.com',
    license='LGPL-3.0',
    license_file='LICENSE',
    classifiers=[
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    ext_modules=[NodeJsExtension('pycover')],
    packages=find_packages(),
    install_requires=[
        'aiohttp',
        'psutil',
        'screeninfo',
    ],
    python_requires='>=3.6.1',
    include_package_data=True,
    universal=True,
    cmdclass={
        'build_ext': NodeJsBuilder,
    },
    zip_safe=False,
)



import asyncio
import logging
import shlex
from typing import Union

from ...exceptions import NodeJSNotRunning
from ...exceptions import NoMtProtoClientSet
from ...exceptions import NotInGroupCallError
from ...file_manager import FileManager
from ...mtproto import BridgedClient
from ...scaffold import Scaffold
from ...types import CaptureAudioDevice
from ...types import CaptureAVDesktop
from ...types import CaptureAVDeviceDesktop
from ...types import CaptureVideoDesktop
from ...types import NotInGroupCall
from ...types import StreamDeleted
from ...types.input_stream import AudioPiped
from ...types.input_stream import AudioVideoPiped
from ...types.input_stream import InputStream
from ...types.input_stream import VideoPiped
from ...types.input_stream.audio_image_piped import AudioImagePiped
from ...types.session import Session

py_logger = logging.getLogger('pytgcalls')


class ChangeStream(Scaffold):
    async def change_stream(
        self,
        chat_id: Union[int, str],
        stream: InputStream,
    ):
        """Change the streaming file

        This method allow to change streaming file
        to a Group Call

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier of the target chat.
                Can be a direct id (int) or a username (str)
            stream (:obj:`~pycover.types.InputStream()`):
                Input Streams descriptor, can be used also
                :obj:`~pycover.types.AudioPiped()`,
                :obj:`~pycover.types.AudioImagePiped()`,
                :obj:`~pycover.types.AudioVideoPiped()` or
                :obj:`~pycover.types.VideoPiped()`

        Raises:
            NoMtProtoClientSet: In case you try
                to call this method without any MtProto client
            NodeJSNotRunning: In case you try
                to call this method without do
                :meth:`~pycover.pycover.start` before
            NoActiveGroupCall: In case you try
                to edit a not started group call
            FileNotFoundError: In case you try
                a non-existent file
            InvalidStreamMode: In case you try
                to set a void stream mode
            FFmpegNotInstalled: In case you try
                to use the Piped input stream, and
                you don't have ffmpeg installed
            NoAudioSourceFound: In case you try
                to play an audio file from a file
                without the sound
            NoVideoSourceFound: In case you try
                to play a video file from a file
                without the video
            InvalidVideoProportion: In case you try
                to play a video without correct
                proportions
            NotInGroupCallError: In case you try
                to leave a non-joined group call

        Example:
            .. code-block:: python
                :emphasize-lines: 10-15

                from pycover import Client
                from pycover import idle
                ...

                app = PyCover(client)
                app.start()

                ...  # Call API methods

                app.change_stream(
                    -1001185324811,
                    AudioPiped(
                        'test.mp4',
                    )
                )

                idle()
        """
        try:
            chat_id = int(chat_id)
        except ValueError:
            chat_id = BridgedClient.chat_id(
                await self._app.resolve_peer(chat_id),
            )
        if self._app is not None:
            if self._wait_until_run is not None:
                headers = None
                if isinstance(
                    stream,
                    AudioImagePiped,
                ) or isinstance(
                    stream,
                    AudioPiped,
                ) or isinstance(
                    stream,
                    AudioVideoPiped,
                ) or isinstance(
                    stream,
                    VideoPiped,
                ):
                    headers = stream.raw_headers
                if stream.stream_video is not None:
                    if not stream.stream_video.path.startswith('screen://'):
                        await FileManager.check_file_exist(
                            stream.stream_video.path.replace(
                                'fifo://',
                                '',
                            ).replace(
                                'image:',
                                '',
                            ),
                            headers,
                        )
                if stream.stream_audio is not None:
                    if not stream.stream_audio.path.startswith('device://'):
                        await FileManager.check_file_exist(
                            stream.stream_audio.path.replace(
                                'fifo://',
                                '',
                            ).replace(
                                'image:',
                                '',
                            ),
                            headers,
                        )
                audio_f_parameters = ''
                video_f_parameters = ''
                if isinstance(
                    stream,
                    AudioImagePiped,
                ) or isinstance(
                    stream,
                    AudioPiped,
                ) or isinstance(
                    stream,
                    AudioVideoPiped,
                ) or isinstance(
                    stream,
                    VideoPiped,
                ) or isinstance(
                    stream,
                    CaptureVideoDesktop,
                ) or isinstance(
                    stream,
                    CaptureAudioDevice,
                ):
                    await stream.check_pipe()
                    if stream.stream_audio:
                        if stream.stream_audio.header_enabled:
                            audio_f_parameters = stream.headers
                    audio_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.ffmpeg_parameters),
                    )
                    if stream.stream_video:
                        if stream.stream_video.header_enabled:
                            video_f_parameters = stream.headers
                    video_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.ffmpeg_parameters),
                    )
                elif isinstance(
                    stream,
                    CaptureAVDeviceDesktop,
                ):
                    audio_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.audio_ffmpeg),
                    )
                    video_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.video_ffmpeg),
                    )
                elif isinstance(
                    stream,
                    CaptureAVDesktop,
                ):
                    await stream.check_pipe()
                    if stream.stream_audio:
                        if stream.stream_audio.header_enabled:
                            audio_f_parameters = stream.headers
                    audio_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.audio_ffmpeg),
                    )
                    video_f_parameters += ':_cmd_:'.join(
                        shlex.split(stream.video_ffmpeg),
                    )
                solver_id = Session.generate_session_id(24)

                async def internal_sender():
                    if not self._wait_until_run.done():
                        await self._wait_until_run
                    stream_audio = stream.stream_audio
                    stream_video = stream.stream_video
                    request = {
                        'action': 'change_stream',
                        'chat_id': chat_id,
                        'lip_sync': stream.lip_sync,
                        'solver_id': solver_id,
                    }
                    if stream_audio is not None:
                        request['stream_audio'] = {
                            'path': stream_audio.path,
                            'bitrate': stream_audio.parameters.bitrate,
                            'ffmpeg_parameters': audio_f_parameters,
                        }
                    if stream.stream_video is not None:
                        video_parameters = stream_video.parameters
                        if video_parameters.frame_rate % 5 != 0 and \
                                not isinstance(stream, AudioImagePiped):
                            py_logger.warning(
                                'For better experience the '
                                'video frame rate must be a multiple of 5',
                            )
                        request['stream_video'] = {
                            'path': stream_video.path,
                            'width': video_parameters.width,
                            'height': video_parameters.height,
                            'framerate': video_parameters.frame_rate,
                            'ffmpeg_parameters': video_f_parameters,
                        }
                    await self._binding.send(request)
                asyncio.ensure_future(internal_sender())
                result = await self._wait_result.wait_future_update(
                    solver_id,
                )
                if isinstance(result, NotInGroupCall):
                    raise NotInGroupCallError()
                elif isinstance(result, StreamDeleted):
                    raise FileNotFoundError()
            else:
                raise NodeJSNotRunning()
        else:
            raise NoMtProtoClientSet()



import asyncio
import atexit
import json
import logging
import os
import signal
import subprocess
import sys
from asyncio import Future
from asyncio.subprocess import Process
from json import JSONDecodeError
from time import time
from typing import Callable
from typing import Dict
from typing import Optional

from .types.session import Session

py_logger = logging.getLogger('pytgcalls')


class Binding:
    def __init__(
        self,
        overload_quiet_mode: bool,
    ):
        self._js_process: Optional[Process] = None
        self._ssid = ''
        self._on_request: Optional[Callable] = None
        self._on_connect: Optional[Callable] = None
        self._last_ping = 0
        self._waiting_ping: Dict[str, Future] = {}
        self._overload_quiet = overload_quiet_mode

        def cleanup():
            async def async_cleanup():
                try:
                    if self._js_process is not None:
                        if not sys.platform.startswith('win'):
                            self._js_process.send_signal(signal.SIGINT)
                            await asyncio.wait_for(
                                self._js_process.communicate(),
                                timeout=3,
                            )
                        else:
                            self._js_process.kill()
                            await self._js_process.communicate()
                except subprocess.TimeoutExpired:
                    py_logger.warning(
                        'Node.js did not terminate cleanly, '
                        'killing process...',
                    )
                    self._js_process.kill()
                    await self._js_process.communicate()
                except ProcessLookupError:
                    pass
                py_logger.info('Node.js stopped')

            asyncio.get_event_loop().run_until_complete(async_cleanup())

        atexit.register(cleanup)

    def on_update(self) -> Callable:
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._on_request = func
            return func

        return decorator

    def on_connect(self):
        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._on_connect = func
            return func

        return decorator

    def is_alive(self):
        return int(time()) - self._last_ping < 15

    @property
    async def ping(self) -> float:
        start_time = time()
        session = Session.generate_session_id(15)
        loop = asyncio.get_event_loop()
        self._waiting_ping[session] = loop.create_future()
        await self._send({
            'ping_with_response': True,
            'sid': session,
        })
        await self._waiting_ping[session]
        del self._waiting_ping[session]
        return (time() - start_time) * 1000.0

    @property
    def _run_folder(self):
        return f'{__file__.replace("binding.py", "")}'

    async def connect(
        self,
        event: Future,
        user_id: int,
    ):
        if self._js_process is None:
            self._js_process = await asyncio.create_subprocess_exec(
                'node',
                os.path.join(self._run_folder, 'dist', 'index.js'),
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )
            event.set_result(None)
            while True:
                try:
                    if self._js_process.stdout is None:
                        break
                    out = (await self._js_process.stdout.readline()) \
                        .decode().replace('\r', '')
                    if not out:
                        break
                    list_data = out.split('\n')
                    for update in list_data:
                        try:
                            json_out = json.loads(update)
                            if 'ping_with_response' in json_out:
                                session_id = json_out['sid']
                                if session_id in self._waiting_ping:
                                    self._waiting_ping[
                                        session_id
                                    ].set_result(
                                        None,
                                    )
                            if 'ping' in json_out:
                                self._last_ping = int(time())
                            if 'try_connect' in json_out:
                                self._ssid = json_out['try_connect']
                                asyncio.ensure_future(
                                    self._send({
                                        'try_connect': 'connected',
                                        'user_id': user_id,
                                        'overload_quiet': self._overload_quiet,
                                    }),
                                )
                                if self._on_connect is not None:
                                    asyncio.ensure_future(self._on_connect())
                            elif 'ssid' in json_out and 'uid' in json_out:
                                if json_out['ssid'] == self._ssid:
                                    if self._on_request is not None:
                                        async def future_response(
                                            future_json_out: dict,
                                        ):
                                            if self._on_request is None:
                                                return
                                            result = await self._on_request(
                                                future_json_out['data'],
                                            )
                                            if isinstance(result, dict):
                                                await self._send_response(
                                                    result,
                                                    future_json_out['uid'],
                                                )
                                            else:
                                                await self._send_error(
                                                    'INVALID_RESPONSE',
                                                    future_json_out['uid'],
                                                )

                                        asyncio.ensure_future(
                                            future_response(json_out),
                                        )
                            elif 'log_message' in json_out \
                                    and 'verbose_mode' in json_out:
                                if json_out['verbose_mode'] == 1:
                                    py_logger.debug(json_out['log_message'])
                                elif json_out['verbose_mode'] == 2:
                                    py_logger.info(json_out['log_message'])
                                elif json_out['verbose_mode'] == 3:
                                    py_logger.warning(json_out['log_message'])
                                elif json_out['verbose_mode'] == 4:
                                    py_logger.error(json_out['log_message'])
                        except JSONDecodeError:
                            if update:
                                if ':replace_line:' in update:
                                    print(
                                        update.replace(
                                            ':replace_line:', '',
                                        ),
                                        end='\r',
                                    )
                                else:
                                    print(update)
                except TimeoutError:
                    pass

    async def _send_response(self, json_data: dict, uid: str):
        if self._ssid:
            await self._send({
                'data': json_data,
                'uid': uid,
                'ssid': self._ssid,
            })

    async def _send_error(self, err_mess: str, uid: str):
        if self._ssid:
            await self._send({
                'err_mess': err_mess,
                'uid': uid,
                'ssid': self._ssid,
            })

    async def send(self, json_data: dict):
        await self._send({
            'data': json_data,
            'ssid': self._ssid,
        })

    async def _send(self, json_data: dict):
        try:
            if self._js_process is not None:
                if self._js_process.stdin is not None:
                    self._js_process.stdin.write(
                        json.dumps(json_data).encode(),
                    )
                    await self._js_process.stdin.drain()
        except ConnectionResetError:
            pass



import asyncio

from ...pycover_session import PyCoverSession
from ...scaffold import Scaffold


class BindingRunner(Scaffold):
    async def _start_binding(self):
        @self._binding.on_update()
        async def update_handler(data: dict):
            if 'action' in data:
                if data['action'] == 'join_voice_call_request':
                    return await self._join_voice_call(data['payload'])
                elif data['action'] == 'leave_call_request':
                    return await self._leave_voice_call(data)
                elif data['action'] == 'stream_audio_ended':
                    return await self._stream_ended_handler(data, True)
                elif data['action'] == 'stream_video_ended':
                    return await self._stream_ended_handler(data, False)
                elif data['action'] == 'upgrade_video_status':
                    return await self._set_video_call_status(data)
                elif data['action'] == 'update_request':
                    return await self._raw_update_handler(data)
            return {
                'result': 'INVALID_REQUEST',
            }

        @self._binding.on_connect()
        async def connect():
            try:
                self._wait_until_run.set_result(None)
            except Exception as e:
                print(e)

            pass

        await PyCoverSession().start()
        loop = asyncio.get_event_loop()
        started_core = loop.create_future()
        self._async_core = asyncio.ensure_future(
            self._binding.connect(
                started_core,
                self._my_id,
            ),
        )
        try:
            await started_core
        except KeyboardInterrupt:
            pass



from ..audio_parameters import AudioParameters


class MediumQualityAudio(AudioParameters):
    """Medium Audio Quality (36K of bitrate)

    Attributes:
        bitrate (``int``):
            Audio bitrate
    """

    def __init__(self):
        super().__init__(
            36000,
        )



from typing import Callable

from ...scaffold import Scaffold


class OnStreamEnd(Scaffold):
    def on_stream_end(self) -> Callable:
        """Decorator for handling when a stream playing
        is ended

        When a streaming will end, this decorator will
        be raised

        Example:
            .. code-block:: python
                :emphasize-lines: 4-5

                ...
                app = PyCover(client)
                ...
                @app.on_stream_end()
                async def handler(client: PyCover, update: Update):
                    print(update)
                ...
                app.run()

        """

        method = 'STREAM_END_HANDLER'

        def decorator(func: Callable) -> Callable:
            if self is not None:
                self._on_event_update.add_handler(
                    method,
                    func,
                )
            return func
        return decorator
